import { Component, OnInit, OnDestroy, ChangeDetectorRef, AfterContentInit, AfterViewInit } from '@angular/core';

import { Store, select } from '@ngrx/store';

import { of, Subscription } from 'rxjs';
import { map, catchError, take } from 'rxjs/operators';

import { TranslateService } from '@ngx-translate/core';

import { ElectronProxyService } from './service/electron-proxy.service';
import { MenuEvent } from '../commons/type';
import * as AppStore from './store';
import * as _AppStore from './store/app';
import * as UserStore from './store/environment/user';

import * as log from './core/util/log';

@Component({
  selector: 'app-root',
  templateUrl: './app.component.html',
  styleUrls: ['./app.component.scss']
})
export class AppComponent implements OnInit, AfterContentInit, AfterViewInit, OnDestroy {
  title = 'scanner-app';
  showTitleBar: boolean;

  menuSubscription: Subscription;
  displayAbout: boolean;
  displayExportJPEG: boolean;
  displayExportPNG: boolean;
  displayExportSVG: boolean;
  displayPreferences: boolean;
  displayPrint: boolean;

  blockAppContent$ = this.store.pipe(select(AppStore.UISelector.LayoutSelector.selectBlockAppContent));

  public constructor(
    private changeDetector: ChangeDetectorRef,
    private store: Store<any>,
    private translateService: TranslateService,
    private electronProxyService: ElectronProxyService,
  ) {
    // this.showTitleBar = !__LINUX__;
    this.showTitleBar = false;

    this.translateService.addLangs(['en', 'ko']);
    this.translateService.setDefaultLang('en');
  }

  ngOnInit(): void {
    const __this = this;

    log.info('AppComponent.ngOnInit');

    this.translateService.use('en').pipe(take(1)).subscribe((res: any) => {
      __this.changeDetector.detectChanges();
    });

    this.store.dispatch(new _AppStore.AppInit());
    this.store.dispatch(new UserStore.SetMemberID({ memberID: 'scannerUser' }));
    // this.probeService.connect();
    this.electronProxyService.sendReady(performance.now());
    this.menuSubscription = this.electronProxyService.menuObservable()
      .pipe(
        map((name: MenuEvent) => {
          let requireDetectChanges = true;
          switch (name) {
            case 'show-about':
              this.displayAbout = true;
              break;
            case 'show-export-jpeg':
              this.displayExportJPEG = true;
              break;
            case 'show-export-png':
              this.displayExportPNG = true;
              break;
            case 'show-export-svg':
              this.displayExportSVG = true;
              break;
            case 'show-preferences':
              this.displayPreferences = true;
              break;
            case 'show-print':
              this.displayPrint = true;
              break;
            case 'change-language-english':
              requireDetectChanges = false;
              this.translateService.use('en').pipe(take(1)).subscribe((res: any) => {
                __this.changeDetector.detectChanges();
              });
              break;
            case 'change-language-korean':
              requireDetectChanges = false;
              this.translateService.use('ko').pipe(take(1)).subscribe((res: any) => {
                __this.changeDetector.detectChanges();
              });
              break;
            default:
              break;
          }
          if (requireDetectChanges) {
            __this.changeDetector.detectChanges();
          }
        }),
        catchError(error => {
          return of(error);
        }),
      )
      .subscribe();

  }

  ngAfterContentInit(): void {
    this.store.dispatch(new _AppStore.AppAfterContentInit());

  }

  ngAfterViewInit(): void {
    this.store.dispatch(new _AppStore.AppAfterViewInit());
  }


  ngOnDestroy(): void {
    this.store.dispatch(new _AppStore.AppDestroy());

    this.menuSubscription.unsubscribe();
  }

  // onMenuClosed(event) {
  //   this.changeDetector.markForCheck();
  // }
}
