import {
  Component,
  Input,
  OnInit,
  ViewChild,
  OnDestroy,
  ChangeDetectorRef
} from '@angular/core';

import { Store, select } from '@ngrx/store';

import { Observable, of, Subscription } from 'rxjs';
import { catchError, map, tap, exhaustMap } from 'rxjs/operators';

import { SelectItem } from 'primeng/primeng';

import { Interface } from '@overflow/model/net/nic';
import { Zone } from '@overflow/model/discovery';
import { toMetaIPType, MetaIPType, MetaIPTypeEnum } from '@overflow/model/meta';
import { DropdownPanelComponent } from '@overflow/commons/ui/component/primeng';

import * as AppStore from '../../../store';
import * as DiscoveryConfigStore from '../../../store/discovery/config';

import * as log from '../../../core/util/log';

class NICInfo {
  iface: string;
  friendlyName: string;
  metaIPType: MetaIPType;
  mac: string;
  address: string;
  network: string;
}

@Component({
  selector: 'app-layout-toolbar-nic-dropdown',
  templateUrl: './nic-dropdown.component.html',
  styleUrls: ['./nic-dropdown.component.scss']
})
export class NicDropdownComponent implements OnInit, OnDestroy {

  @Input()
  blockTarget: any;

  @ViewChild('panel')
  panel: DropdownPanelComponent;

  pending$: Observable<boolean>;
  error$: Observable<any>;

  interfacesSubscription: Subscription | null;

  nics: SelectItem[] | null;
  selected: NICInfo;

  constructor(
    private store: Store<any>,
    private changeDetector: ChangeDetectorRef,
  ) { }

  ngOnInit(): void {
    const __this = this;
    this.interfacesSubscription = this.store.pipe(
      tap(() => {
        __this.pending$ = of(true);
      }),
      select(AppStore.EnvironmentSelector.InterfacesSelector.selectInterfaces),
      map((ifaces: Interface[]) => {
        if (null === ifaces) {
          return;
        }
        __this.nics = [];
        ifaces.forEach(iface => {
          if (undefined === iface.addresses) {
            return;
          }
          iface.addresses.forEach(address => {
            if (
              address.metaIPType.key !== toMetaIPType(MetaIPTypeEnum.V4).key
            ) {
              return;
            }
            const nicInfo = {
              iface: iface.iface,
              friendlyName: iface.friendlyName,
              metaIPType: address.metaIPType,
              mac: iface.mac,
              address: address.address,
              network: address.network
            };
            __this.nics.push({
              label: iface.friendlyName + ' (' + address.network + ')',
              value: nicInfo,
              disabled:
                address.metaIPType.key !== toMetaIPType(MetaIPTypeEnum.V4).key
                  ? true
                  : false
            });
            if (address.gateway !== undefined && address.gateway.length > 0) {
              __this.nicSelected(nicInfo);
            }
          });
        });
        this.changeDetector.detectChanges();
      }),
      catchError(error => {
        log.error(error);
        return of();
      }),
      tap(() => {
        __this.pending$ = of(false);
      }),
    ).subscribe();
  }

  ngOnDestroy(): void {
    if (null !== this.interfacesSubscription) {
      this.interfacesSubscription.unsubscribe();
    }
  }

  nicSelected(nic: NICInfo) {
    this.selected = nic;

    const zone: Zone = {
      network: this.selected.network,
      iface: this.selected.iface,
      metaIPType: toMetaIPType(MetaIPTypeEnum.V4),
      address: this.selected.address,
      mac: this.selected.mac
    };

    this.store.dispatch(new DiscoveryConfigStore.ChangeZone(zone));

    this.panel.hide();
  }
}
